.docItemContainer {
    margin-bottom: 50px;
}

.docItemContainer header + *,
.docItemContainer article > *:first-child {
  margin-top: 0;
}

@media (min-width: 997px) {
  .docItemCol {
    max-width: 75% !important;
  }
}
