.docItemContent {
    display: flex;
    align-items: center;
    justify-content: space-between;
    gap: 1rem;
    flex-wrap: wrap;
    padding-bottom: calc(
        var(--ifm-h1-vertical-rhythm-bottom) * var(--ifm-leading)
    );
  
    h1 {
        margin: 0 !important;
        flex: 1 1 auto;
        min-width: 12rem;
    }
  
    @media (max-width: 767px) {
        flex-direction: column;
        align-items: flex-start;
        gap: 0.75rem;
    }
  }