.button {
    display: inline-flex;
    align-items: center;
    text-align: center;
    padding: 8px 16px;
    border-radius: 8px;
    font-family: (--ifm-font-family-base);
    font-size: 16px;
    font-style: normal;
    font-weight: 500;
    line-height: 24px;
    cursor: pointer;
    transition: background-color 0.2s;

    svg {
        margin-right: 8px;
    }
}

.buttonPrimary {
    background-color: var(--color-black-action);
    color: var(--color-text-on-primary);
    border: none;

    path {
        stroke: var(--color-text-on-primary);
        &:first-child {
            fill: var(--color-text-on-primary);
        }
    }
}

.buttonPrimary:hover {
    background-color: var(--color-primary-action-hover);
}

.buttonSecondary {
    background-color: var(--color-background);
    color: var(--color-text);
    border: 1px solid var(--color-border);

    path {
        stroke: var(--color-black-action);
        &:first-child {
            fill: var(--color-black-action);
        }
    }
}

.buttonSecondary:hover {
    border: 1px solid var(--color-text);
}

.big {
    padding: 12px 24px;
}

/* TABLET */
@media (min-width: 768px) {
    .button {
        width: auto;
    }
}
