from __future__ import annotations

COMMON_ACCEPT_LANGUAGE = 'en-US,en;q=0.9'

BROWSER_TYPE_HEADER_KEYWORD = {
    'chrome': {'Chrome', 'CriOS'},
    'firefox': {'Firefox', 'FxiOS'},
    'edge': {'Edg', 'Edge', 'EdgA', 'EdgiOS'},
    'safari': {'Safari'},
}
