.footer {
    background: var(--color-background);
    color: var(--color-text);
}

.footerBottom,
.footerTop {
    border-top: 1px solid var(--color-separator);

    @media (min-width: 768px) {
        padding: 40px 40px;
    }
    @media (min-width: 1024px) {
        padding: 40px 64px;
    }
}

.footerTopRow {
    max-width: var(--max-layout-width);
    margin: auto;

    display: flex;
    flex-direction: column;
    @media (min-width: 768px) {
        flex-direction: row;
    }
}

.footerTopRowRight {
    flex-direction: column;
    display: flex;
    flex: 3;
    gap: 32px;
    padding: 16px 40px 40px;

    @media (min-width: 768px) {
        gap: 0;
        flex-direction: row;
        padding: 0;
        justify-content: space-between;
    }
}

.footerTopRowLeft {
    display: flex;
    flex-direction: column;
    justify-content: space-between;
    flex: 2;
    padding: 32px 40px 24px;
    gap: 32px;

    border-bottom: 1px solid var(--color-separator);

    img {
        display: block !important;
    }

    @media (min-width: 768px) {
        padding: 0;
        border: 0;
        gap: 0;
    }
}

.footerBottomRow {
    max-width: var(--max-layout-width);
    margin: auto;

    display: flex;
    flex-direction: column;
    align-items: center;
    gap: 24px;
    padding: 24px 40px;

    font-size: 14px;
    line-height: 20px;
    text-align: center;

    @media (min-width: 768px) {
        gap: 0;
        padding: 0;
        flex-direction: row;
        justify-content: space-between;
    }
}

.hearthIcon {
    margin-right: 8px;
    path {
        fill: var(--color-text-muted);
    }
}

.footerTitle {
    font-size: 16px;
    font-weight: 700;
    line-height: 24px;
}

.footerList {
    margin: 0;
    li {
        margin-top: 16px;
        height: 28px;
    }
}

.footerLogo {
    width: fit-content;
}
