.toggleButton {
    padding: 4px;
    display: flex;
    gap: 4px;
    align-items: center;
    transition: all var(--ifm-transition-fast);
    position: relative;
    border-radius: 150px;
    background-color: var(--color-background-subtle);
}

.toggleButton span {
    width: 44px;
    height: 36px;
    border-radius: 50%;
    background: #fff;
    position: absolute;
    transition: all var(--ifm-transition-fast);
    left: 0;
    margin: 4px;

    border-radius: 150px;
    background-color: var(--color-background);

    /* Light/L1 */
    box-shadow:
        0px 0.5px 1.5px 0px rgba(63, 71, 93, 0.15),
        0.4px 0.8px 1px -1.2px rgba(63, 71, 93, 0.14),
        1px 2px 2.5px -2.5px rgba(63, 71, 93, 0.13);
}

.toggleButton svg {
    z-index: 1;
    margin: 8px 12px;
    width: 20px;
    height: 20px;
    path {
        stroke: var(--color-icon);
    }
}

[data-theme='dark'] .toggleButton span {
    left: 48px;
}

.toggleButtonDisabled {
    cursor: not-allowed;
}
