.navbarItems {
    display: flex;
    align-items: center;
    margin-inline: auto;
    gap: 16px;
}

.navbarItems__leftMargin {
    margin-left: 40px;
}

.getStartedButton {
    color: var(--color-text-on-primary);
    background: var(--color-black-action);
    border-radius: 8px;
    font-size: 16px;
    font-weight: 500;
    line-height: 24px;
    padding: 8px 16px !important;
    border: none;
    transition: background-color 0.2s;

    &:hover {
        color: var(--color-text-on-primary);
        background-color: var(--color-primary-action-hover);
    }
}
