.languageGetStartedContainer {
    margin: 0;
    display: flex;
    flex-direction: column;
    align-items: center;
    padding-inline: 12px;
}

.languageGetStartedContainer img {
    height: 40px;
    margin-bottom: 16px;
}

.buttonContainer {
    display: flex;
    flex-direction: column;
    align-items: center;
    gap: 16px;
    & > span {
        line-height: 0;
        min-height: 28px;
    }
    a,
    a span {
        min-width: 190px;
        text-align: center;
        justify-content: center;
    }
}

.buttonContainer:has(+ code) {
    margin-bottom: 16px;
    gap: 12px;
}

.commandContainer {
    margin: 0;
    padding: 0;
    color: var(--color-text);
    font-size: 12px;
    font-style: normal;
    font-weight: 400;
    line-height: 16px;
    background-color: transparent;
    border: 0;
    display: flex;
    align-items: center;
}

.commandContainer button {
    opacity: 0;
    transition: opacity var(--ifm-transition-fast) ease-in;
}

.commandContainer:hover button,
.commandContainer button:hover {
    opacity: 1;
}

/* TABLET */
@media (min-width: 768px) {
    .languageGetStartedContainer {
        margin: 24px 0 40px 0;
    }
    .buttonContainer:has(+ code) {
        flex-direction: row;
    }
    .buttonContainer:has(+ code) {
        a,
        a span {
            min-width: 0;
        }
    }
}
